import { __awaiter, __decorate } from "tslib";
import { createLogger } from '@genesislcap/foundation-logger';
import { BaseDatasource } from '@genesislcap/foundation-ui';
import { LifecycleMixin } from '@genesislcap/foundation-utils';
import { attr, customElement } from '@microsoft/fast-element';
import { chartDatasourceEventNames } from './chart-datasource.config';
import { extractFields, getFormattedDateGroupBy, getReduced } from './chart-datasource.utils';
const logger = createLogger('chart-datasource');
/**
 * The first and second default chart fields when it's a dual axes chart.
 * @remarks This field is always used. Default value is 'groupBy value'.
 * @public
 */
export const defaultChartFields = 'groupBy value';
/**
 * The third default chart field when it's a triple axes chart.
 * @remarks This field is optional. Default value is 'series'.
 * @public
 */
export const defaultSeriesChartField = 'series';
const defaultConfigFieldsCount = 2;
const maxConfigFieldsCount = 3;
const groupByConfigFieldIndex = 0;
const valueConfigFieldIndex = 1;
const seriesConfigFieldIndex = 2;
class ConcreteDatasource extends BaseDatasource {
    constructor() {
        super(...arguments);
        this.showFormattedTime = false;
        this.reduced = true;
    }
    get groupByServerFieldName() {
        return this.extractedServerFields[groupByConfigFieldIndex];
    }
    // ChartDatasource's implementation of BaseDatasource abstracts
    clearData() {
        var _a;
        (_a = this.dataSub) === null || _a === void 0 ? void 0 : _a.unsubscribe();
        this.rowsSyncedWithStream = new Map();
        this.chart.data = [];
    }
    syncComponentData(snapshotData) {
        const rowData = snapshotData !== null && snapshotData !== void 0 ? snapshotData : Array.from(this.rowsSyncedWithStream.values());
        let chartData = [];
        rowData.forEach((row) => {
            const chartRow = {};
            this.extractedServerFields.forEach((fieldName, index) => {
                if (this.withTimestampFormatting && fieldName === this.groupByServerFieldName) {
                    row[fieldName] = getFormattedDateGroupBy(row[fieldName], this.validServerFieldsMetadata.get(fieldName).type, this.showFormattedTime);
                }
                chartRow[this.extractedChartsFields[index]] = row[fieldName];
            });
            chartData.push(chartRow);
        });
        if (this.reduced) {
            chartData = getReduced(chartData, this.extractedChartsFields[groupByConfigFieldIndex], this.extractedChartsFields[valueConfigFieldIndex], this.seriesChartField);
        }
        if (this.chart.type === 'dualaxes') {
            this.chart.data = [chartData, chartData];
        }
        else {
            this.chart.data = chartData;
        }
        this.requiresFullDataRefresh = false;
    }
    handleStreamInserts(insertedRows) {
        insertedRows === null || insertedRows === void 0 ? void 0 : insertedRows.forEach((insertData) => {
            this.rowsSyncedWithStream.set(insertData[this.rowId], insertData);
        });
        this.syncComponentData();
    }
    handleStreamDeletes(deletedRows) {
        deletedRows === null || deletedRows === void 0 ? void 0 : deletedRows.forEach((deleteData) => {
            this.rowsSyncedWithStream.delete(deleteData[this.rowId]);
        });
        this.syncComponentData();
    }
    handleStreamUpdates(updatedRows) {
        updatedRows === null || updatedRows === void 0 ? void 0 : updatedRows.forEach((updateData) => {
            const rowToBeUpdated = this.rowsSyncedWithStream.get(updateData[this.rowId]);
            const updatedRow = Object.assign(Object.assign({}, rowToBeUpdated), updateData);
            this.rowsSyncedWithStream.set(updateData[this.rowId], updatedRow);
        });
        this.syncComponentData();
    }
}
/**
 * The Genesis Datasource for the `g2plot-chart` component.
 * @public
 */
let ChartDatasource = class ChartDatasource extends LifecycleMixin(ConcreteDatasource) {
    constructor() {
        super(...arguments);
        /**
         * Default value is 'groupBy value'. Can be customized but order matters.
         *
         * [Mandatory] First is always the xField/colorField
         * [Mandatory] Second is always the yField/angleField
         * [Optional] Third is always the seriesField
         *
         * @public
         */
        this.chartFields = defaultChartFields;
        /**
         * If true, will enable stacked chart to effectively portray comparisons of total values across several categories.
         * @public
         */
        this.isStack = false;
        /**
         * If true, will enable grouped chart plotting numeric values for levels of two categorical variables instead of one.
         * @public
         */
        this.isGroup = false;
        /**
         * Enabled by default, will not work if a custom charts-fields is used (requires [type, value]).
         * @public
         */
        this.reduced = true;
        /**
         * Only works if [withTimestampFormatting] is true.
         *
         * Disabled by default, will ignore the TIME portion in the Chart labels.
         * Only affects [DATETIME] fields.
         * @public
         */
        this.showFormattedTime = false;
    }
    /**
     * @internal
     */
    connectedCallback() {
        const _super = Object.create(null, {
            connectedCallback: { get: () => super.connectedCallback }
        });
        return __awaiter(this, void 0, void 0, function* () {
            _super.connectedCallback.call(this);
            this.chart = this.parentElement;
            this.extractedServerFields = extractFields(this.serverFields);
            this.extractedChartsFields = extractFields(this.chartFields);
            this.prepareSeriesField();
            if (!this.hasValidServerFields || !this.hasValidChartFields) {
                this.handleError({
                    message: 'Use valid server and chart fields (e.g server-fields="NAME VALUE" charts-fields="groupBy value"). Chart will not load.',
                });
                return;
            }
            yield this.initDatasource();
            const invalidServerFieldsFound = this.prepareValidServerFieldsMetadata();
            if (invalidServerFieldsFound.length > 0) {
                const allPossibleValidServerFields = this.datasource.fieldMetadata.map((field) => field.name);
                const allInvalidServerFields = invalidServerFieldsFound.join(' ');
                const message = invalidServerFieldsFound.length > 1
                    ? `Server fields '${allInvalidServerFields}' are invalid.`
                    : `Server field '${allInvalidServerFields}' is invalid.`;
                const filedsNames = allPossibleValidServerFields.join(',');
                this.handleError({
                    message: `${message} Chart will not load. Available fields: ${filedsNames}`,
                });
                return;
            }
            if (this.chart && this.resourceName && this.connect.isConnected) {
                this.fetchGenesisData(false);
            }
        });
    }
    /**
     * @internal
     */
    handleError(params) {
        const { message, loggerType = 'error' } = params;
        if (message) {
            logger[loggerType](message);
            if (loggerType === 'error') {
                this.$emit(chartDatasourceEventNames.error, { message });
            }
        }
    }
    /**
     * @internal
     */
    disconnectedCallback() {
        if (!this.shouldRunDisconnect)
            return;
        super.disconnectedCallback();
    }
    // ChartDatasource's specifics
    /**
     * @internal
     */
    fetchGenesisData(withFullInit) {
        if (!this.shouldRunConnect)
            return;
        return super.fetchGenesisData(withFullInit);
    }
    prepareSeriesField() {
        if (this.extractedServerFields.length === maxConfigFieldsCount &&
            this.extractedChartsFields.length === defaultConfigFieldsCount) {
            this.extractedChartsFields.push(defaultSeriesChartField);
        }
        if (this.extractedChartsFields.length === maxConfigFieldsCount) {
            this.seriesChartField = this.extractedChartsFields[seriesConfigFieldIndex];
        }
        if (this.seriesChartField) {
            this.chart.config = Object.assign(Object.assign({}, this.chart.config), { seriesField: this.seriesChartField });
            switch (this.chart.type) {
                case 'bar':
                case 'column':
                    if (this.isStack && this.isGroup) {
                        this.handleError({
                            message: 'Charts can not be stacked and grouped at the same time. Chart will not load properly.',
                        });
                    }
                    this.chart.config = Object.assign(Object.assign({}, this.chart.config), { isStack: this.isStack, isGroup: this.isGroup });
                    break;
            }
        }
    }
    prepareValidServerFieldsMetadata() {
        const invalidServerFieldNamesFound = [];
        this.validServerFieldsMetadata = new Map();
        this.extractedServerFields.forEach((fieldName) => {
            const fieldMetadata = this.datasource.fieldMetadata.find((fm) => fm.name === fieldName);
            if (fieldMetadata) {
                this.validServerFieldsMetadata.set(fieldName, fieldMetadata);
            }
            else {
                invalidServerFieldNamesFound.push(fieldName);
            }
        });
        return invalidServerFieldNamesFound;
    }
    /**
     * @internal
     */
    get hasValidServerFields() {
        return this.hasValidFieldsStructure(this.extractedServerFields);
    }
    /**
     * @internal
     */
    get hasValidChartFields() {
        return this.hasValidFieldsStructure(this.extractedChartsFields);
    }
    hasValidFieldsStructure(fieldsArray) {
        return (fieldsArray &&
            fieldsArray.length >= defaultConfigFieldsCount &&
            fieldsArray.length <= maxConfigFieldsCount);
    }
};
__decorate([
    attr({ attribute: 'chart-fields' })
], ChartDatasource.prototype, "chartFields", void 0);
__decorate([
    attr({ attribute: 'server-fields' })
], ChartDatasource.prototype, "serverFields", void 0);
__decorate([
    attr({ mode: 'boolean', attribute: 'is-stack' })
], ChartDatasource.prototype, "isStack", void 0);
__decorate([
    attr({ mode: 'boolean', attribute: 'is-group' })
], ChartDatasource.prototype, "isGroup", void 0);
__decorate([
    attr({ mode: 'boolean' })
], ChartDatasource.prototype, "reduced", void 0);
__decorate([
    attr({ mode: 'boolean', attribute: 'show-formatted-time' })
], ChartDatasource.prototype, "showFormattedTime", void 0);
ChartDatasource = __decorate([
    customElement({
        name: 'chart-datasource',
    })
], ChartDatasource);
export { ChartDatasource };
